@@ -67,10 +67,11 @@ gtk_html_search_dialog_new (GtkHTML *html)
 	GtkWidget *hbox;
 
 	dialog->dialog         = GNOME_DIALOG (gnome_dialog_new (NULL, _("Find"), GNOME_STOCK_BUTTON_CANCEL, NULL));
-	dialog->entry          = gtk_entry_new_with_max_length (20);
+	dialog->entry          = gtk_entry_new ();
 	dialog->backward       = gtk_check_button_new_with_label (_("backward"));
 	dialog->case_sensitive = gtk_check_button_new_with_label (_("case sensitive"));
 	dialog->html           = html;
+	dialog->regular        = FALSE;
 
 	hbox = gtk_hbox_new (FALSE, 0);
 
@@ -106,11 +107,11 @@ gtk_html_search_dialog_destroy (GtkHTMLSearchDialog *d)
 void
 search (GtkHTMLControlData *cd, gboolean regular)
 {
+	RUN_DIALOG (search, regular ? _("Find Regular Expression") :  _("Find"));
+
 	if (cd->search_dialog)
 		cd->search_dialog->regular = regular;
 
-	RUN_DIALOG (search, regular ? _("Find Regular Expression") :  _("Find"));
-
 	if (cd->search_dialog)
 		gtk_widget_grab_focus (cd->search_dialog->entry);
 }
