@@ -28,6 +28,7 @@
 #include <config.h>
 #include <ctype.h>
 #include <string.h>
+#include <stdlib.h>
 
 #include <gal/widgets/e-unicode.h>
 
@@ -44,11 +45,12 @@
 #include "htmllinktext.h"
 #include "htmlpainter.h"
 #include "htmlplainpainter.h"
+#include "htmlsearch.h"
+#include "htmlselection.h"
 #include "htmltable.h"
 #include "htmltablecell.h"
 #include "htmltext.h"
 #include "htmltextslave.h"	/* FIXME */
-#include "htmlsearch.h"
 #include "htmlvspace.h"
 
 
@@ -155,18 +157,30 @@ get_recursive_length (HTMLObject *self)
 static HTMLObject *
 op_helper (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, GList *left, GList *right, guint *len, gboolean cut)
 {
+	HTMLObject *o;
+
 	/* if (!from && to && HTML_IS_TABLE (to->data) && to->next && GPOINTER_TO_INT (to->next->data) == 0)
 		return NULL;
 	if (!to && from && HTML_IS_TABLE (from->data) && from->next && GPOINTER_TO_INT (from->next->data) == 1)
 		return NULL; 
 	if (!from && (*len || !(self->prev && HTML_IS_CLUEFLOW (self->prev) && HTML_IS_TABLE (HTML_CLUE (self->prev)->tail))))
 	(*len) ++; */
-	if (!from)
+	if (!from && self->prev) {
 		(*len) ++;
-	html_clueflow_remove_text_slaves (HTML_CLUEFLOW (self));
-	return cut
+		/* if (cut)
+		   e->cursor->position --; */
+	}
+	if (cut)
+		html_clueflow_remove_text_slaves (HTML_CLUEFLOW (self));
+	o = cut
 		? (*HTML_OBJECT_CLASS (parent_class)->op_cut) (self, e, from, to, left, right, len)
 		: (*HTML_OBJECT_CLASS (parent_class)->op_copy) (self, e, from, to, len);
+
+	if (!cut && o) {
+		html_clueflow_remove_text_slaves (HTML_CLUEFLOW (o));
+	}
+
+	return o;
 }
 
 static HTMLObject *
@@ -199,25 +213,44 @@ op_cut (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, GList *left, GL
 	return rv;
 }
 
+static void
+set_head_size (HTMLObject *o)
+{
+	if (o && HTML_CLUE (o)->head)
+		HTML_CLUE (o)->head->change |= HTML_CHANGE_SIZE;
+}
+
+static void
+set_tail_size (HTMLObject *o)
+{
+	if (o && HTML_CLUE (o)->tail) {
+		HTML_CLUE (o)->tail->change |= HTML_CHANGE_SIZE;
+	}
+}
+
+static void
+set_around_size (HTMLObject *o) {
+	if (o) {
+		o->change |= HTML_CHANGE_SIZE;
+		if (o->next)
+			o->next->change |= HTML_CHANGE_SIZE;
+		if (o->prev)
+			o->prev->change |= HTML_CHANGE_SIZE;
+	}
+}
+
 static void
 split (HTMLObject *self, HTMLEngine *e, HTMLObject *child, gint offset, gint level, GList **left, GList **right)
 {
+	set_around_size (child);
 	html_clueflow_remove_text_slaves (HTML_CLUEFLOW (self));
 	(*HTML_OBJECT_CLASS (parent_class)->split) (self, e, child, offset, level, left, right);
+
 	update_item_number (self);
 }
 
-inline static gboolean
-could_merge (HTMLClueFlow *cf1, HTMLClueFlow *cf2)
-{
-	return ((HTML_IS_TABLE (HTML_CLUE (cf1)->tail) && !html_clueflow_is_empty (cf2))
-		|| (HTML_IS_TABLE (HTML_CLUE (cf2)->head) && !html_clueflow_is_empty (cf1)))
-		? FALSE
-		: TRUE;
-}
-
 static gboolean
-merge (HTMLObject *self, HTMLObject *with, HTMLEngine *e, GList *left, GList *right)
+merge (HTMLObject *self, HTMLObject *with, HTMLEngine *e, GList **left, GList **right, HTMLCursor *cursor)
 {
 	HTMLClueFlow *cf1, *cf2;
 	gboolean rv;
@@ -228,25 +261,27 @@ merge (HTMLObject *self, HTMLObject *with, HTMLEngine *e, GList *left, GList *ri
 	html_clueflow_remove_text_slaves (cf1);
 	html_clueflow_remove_text_slaves (cf2);
 
+	set_tail_size (self);
+	set_head_size (with);
+
 	/* printf ("merge flows\n"); */
 
-	if (!could_merge (cf1, cf2)) {
-		update_item_number (self);
-		return FALSE;
-	}
-
-	/* printf ("merge happens\n"); */
 	if (html_clueflow_is_empty (cf1)) {
 		cf1->style = cf2->style;
 		cf1->level = cf2->level;
 		cf1->item_type = cf2->item_type;
 		cf1->item_number = cf2->item_number - 1;
+		self->x = with->x;
+		self->y = with->y;
+		self->width = with->width;
+		self->ascent = with->ascent;
+		self->descent = with->descent;
 		HTML_CLUE  (cf1)->halign = HTML_CLUE (cf2)->halign;
 		HTML_CLUE  (cf1)->valign = HTML_CLUE (cf2)->valign;
 		html_object_copy_data_from_object (self, with);
 	}
 
-	rv = (* HTML_OBJECT_CLASS (parent_class)->merge) (self, with, e, left, right);
+	rv = (* HTML_OBJECT_CLASS (parent_class)->merge) (self, with, e, left, right, cursor);
 
 	update_item_number (self);
 	cf1->item_number --;
@@ -313,9 +348,10 @@ is_header (HTMLClueFlow *flow)
 	}
 }
 
+static guint get_post_padding (HTMLClueFlow *flow, guint pad);
+
 static guint
-get_pre_padding (HTMLClueFlow *flow,
-		 guint pad)
+get_pre_padding (HTMLClueFlow *flow, guint pad)
 {
 	HTMLObject *prev_object;
 
@@ -326,8 +362,21 @@ get_pre_padding (HTMLClueFlow *flow,
 	if (HTML_OBJECT_TYPE (prev_object) == HTML_TYPE_CLUEFLOW) {
 		HTMLClueFlow *prev;
 
+		if (get_post_padding (HTML_CLUEFLOW (prev_object), 1))
+			return 0;
+
+		if (is_item (HTML_CLUEFLOW (prev_object))) {
+			if (is_item (flow)) {
+				return 0;
+			} else {
+				return pad;
+			}
+		} else if (is_item (flow)) {
+			return pad;
+		}
+
 		prev = HTML_CLUEFLOW (prev_object);
-		if (prev->level > 0 && flow->level == 0)
+		if (prev->level > flow->level)
 			return pad;
 
 		if (flow->style == HTML_CLUEFLOW_STYLE_PRE
@@ -360,8 +409,11 @@ get_post_padding (HTMLClueFlow *flow,
 	if (HTML_OBJECT_TYPE (next_object) == HTML_TYPE_CLUEFLOW) {
 		HTMLClueFlow *next;
 
+		if (is_item (flow) && is_item (HTML_CLUEFLOW (next_object)))
+			return 0;
+
 		next = HTML_CLUEFLOW (next_object);
-		if (next->level > 0 && flow->level == 0)
+		if (next->level > flow->level)
 			return pad;
 
 		if (flow->style == HTML_CLUEFLOW_STYLE_PRE
@@ -493,11 +545,28 @@ pref_right_margin (HTMLPainter *p, HTMLClueFlow *clueflow, HTMLObject *o, gint y
 	return MIN (fixed_margin, 72 * html_painter_get_space_width (p, GTK_HTML_FONT_STYLE_SIZE_3, NULL));
 }
 
+static void
+add_clear_area (GList **changed_objs, HTMLObject *o, gint x, gint w)
+{
+	HTMLObjectClearRectangle *cr;
+
+	cr = g_new (HTMLObjectClearRectangle, 1);
+
+	cr->object = o;
+	cr->x = x;
+	cr->y = 0;
+	cr->width = w;
+	cr->height = o->ascent + o->descent;
+
+	*changed_objs = g_list_prepend (*changed_objs, cr);
+	/* NULL meens: clear rectangle follows */
+	*changed_objs = g_list_prepend (*changed_objs, NULL);
+}
+
 /* EP CHECK: should be mostly OK.  */
 /* FIXME: But it's awful.  Too big and ugly.  */
 static gboolean
-calc_size (HTMLObject *o,
-	   HTMLPainter *painter)
+calc_size (HTMLObject *o, HTMLPainter *painter, GList **changed_objs)
 {
 	HTMLVSpace *vspace;
 	HTMLClue *clue;
@@ -512,7 +581,7 @@ calc_size (HTMLObject *o,
 	gint oldy;
 	gint w, a, d;
 	guint padding;
-	gboolean changed;
+	gboolean changed, leaf_childs_changed_size;
 	gint old_ascent, old_descent, old_width;
 	gint runWidth = 0;
 	gboolean have_valign_top;
@@ -520,10 +589,15 @@ calc_size (HTMLObject *o,
 	html_clueflow_remove_text_slaves (HTML_CLUEFLOW (o));
 
 	changed = FALSE;
+	leaf_childs_changed_size = FALSE;
 	old_ascent = o->ascent;
 	old_descent = o->descent;
 	old_width = o->width;
 
+	/* if (changed_objs)
+		printf ("begin:   %d==%d %d==%d %d==%d\n",
+		o->width, old_width, o->ascent, old_ascent, o->descent, old_descent); */
+
 	clue = HTML_CLUE (o);
 	flow = HTML_CLUEFLOW (o);
 
@@ -554,6 +628,11 @@ calc_size (HTMLObject *o,
 	have_valign_top = FALSE;
 
 	while (obj != NULL) {
+
+		if (obj && obj->change & HTML_CHANGE_SIZE
+		    && HTML_OBJECT_TYPE (obj) != HTML_TYPE_TEXTSLAVE && !html_object_is_container (obj))
+			leaf_childs_changed_size = TRUE;
+
 		if (obj->flags & HTML_OBJECT_FLAG_NEWLINE) {
 			if (!a)
 				a = obj->ascent;
@@ -580,7 +659,7 @@ calc_size (HTMLObject *o,
 			HTMLClueAligned *c = (HTMLClueAligned *)obj;
 			
 			if (! html_clue_appended (HTML_CLUE (o->parent), HTML_CLUE (c))) {
-				html_object_calc_size (obj, painter);
+				html_object_calc_size (obj, painter, changed_objs);
 
 				if (HTML_CLUE (c)->halign == HTML_HALIGN_LEFT) {
 					if (obj->x != lmargin) {
@@ -635,7 +714,10 @@ calc_size (HTMLObject *o,
 			if (lmargin >= rmargin) {
 				gint new_y;
 
-				html_object_calc_size (run, painter);
+				if (run && run->change & HTML_CHANGE_SIZE
+				    && HTML_OBJECT_TYPE (run) != HTML_TYPE_TEXTSLAVE && !html_object_is_container (run))
+					leaf_childs_changed_size = TRUE;
+				html_object_calc_size (run, painter, changed_objs);
 				html_clue_find_free_area (HTML_CLUE (o->parent), o->y, run->min_width,
 							  run->ascent + run->descent, indent, &new_y,
 							  &lmargin, &rmargin);
@@ -650,17 +732,27 @@ calc_size (HTMLObject *o,
 				&& ! (run->flags & HTML_OBJECT_FLAG_ALIGNED)) {
 				HTMLFitType fit;
 				HTMLVAlignType valign;
+				gboolean firstRun;
 				gint width_left;
 
+				if (run && run->change & HTML_CHANGE_SIZE
+				    && HTML_OBJECT_TYPE (run) != HTML_TYPE_TEXTSLAVE && !html_object_is_container (run))
+					leaf_childs_changed_size = TRUE;
+
 				width_left = rmargin - runWidth - w;
-				fit = html_object_fit_line (run,
-							    painter,
-							    w + runWidth == lmargin,
-							    run == line || (HTML_OBJECT_TYPE (run) == HTML_TYPE_TEXTSLAVE
-									    && HTML_OBJECT (HTML_TEXT_SLAVE (run)->owner)
-									    == line
-									    && !HTML_TEXT_SLAVE (run)->posStart),
-							    flow->style == HTML_CLUEFLOW_STYLE_PRE ? G_MAXINT : width_left);
+				firstRun = run == line || (HTML_IS_TEXT_SLAVE (run)
+							   && HTML_OBJECT (HTML_TEXT_SLAVE (run)->owner) == line
+							   && HTML_TEXT_SLAVE (run)->posStart == 0);
+
+				if (!firstRun && width_left < 0) {
+					fit = HTML_FIT_NONE;
+				} else {
+					fit = html_object_fit_line (run,
+								    painter,
+								    w + runWidth == lmargin,
+								    firstRun,
+								    flow->style == HTML_CLUEFLOW_STYLE_PRE ? G_MAXINT : width_left);
+				}
 
 				if (fit == HTML_FIT_NONE) {
 					w = set_line_x (&obj, run, w, &changed);
@@ -668,9 +760,8 @@ calc_size (HTMLObject *o,
 					break;
 				}
 
-				html_object_calc_size (run, painter);
+				html_object_calc_size (run, painter, changed_objs);
 				runWidth += run->width;
-
 				valign = html_object_get_valign (run);
 
 				/* Algorithm for dealing vertical alignment.
@@ -782,7 +873,7 @@ calc_size (HTMLObject *o,
 			halign = html_clueflow_get_halignment (flow);
 
 			if (w > o->width)
-				o->width = w; // MIN (w, o->max_width);
+				o->width = w;
 
 			if (halign == HTML_HALIGN_CENTER) {
 				extra = (rmargin - w) / 2;
@@ -891,8 +982,20 @@ calc_size (HTMLObject *o,
 
 	add_post_padding (HTML_CLUEFLOW (o), padding);
 
-	if (o->ascent != old_ascent || o->descent != old_descent || o->width != old_width)
+	if (o->ascent != old_ascent || o->descent != old_descent || o->width != old_width) {
 		changed = TRUE;
+	}
+	if (o->ascent != old_ascent || o->descent != old_descent || o->width != old_width || leaf_childs_changed_size) {
+		/* if (changed_objs)
+			printf ("changed: %d==%d %d==%d %d==%d %d\n",
+			o->width, old_width, o->ascent, old_ascent, o->descent, old_descent, leaf_childs_changed_size); */
+		if (changed_objs) {
+			if (old_width > o->max_width && o->width < old_width) {
+				add_clear_area (changed_objs, o, o->width, old_width - o->width);
+			}
+			html_object_add_to_changed (changed_objs, o);
+		}
+	}
 
 	return changed;
 }
@@ -1083,14 +1186,6 @@ draw_background (HTMLObject *self,
 		 gint width, gint height,
 		 gint tx, gint ty)
 {
-	
-	html_object_draw (self->parent, p,
-			  x + self->parent->x,
-			  y + self->parent->y - self->parent->ascent,
-			  width, height,
-			  tx - self->parent->x,
-			  ty - self->parent->y + self->parent->ascent);
-	
 	html_object_draw_background (self->parent, p,
 				     x + self->parent->x,
 				     y + self->parent->y - self->parent->ascent,
@@ -1210,9 +1305,15 @@ append_selection_string (HTMLObject *self,
 /* Saving support.  */
 
 static gboolean
-write_indent (HTMLEngineSaveState *state)
+write_indent (HTMLEngineSaveState *state, gint level)
 {
-	return html_engine_save_output_string (state, "    ");
+	while (level > 0) {
+		if (!html_engine_save_output_string (state, "    "))
+			return FALSE;
+		level --;
+	}
+
+	return TRUE;
 }
 
 /* static const char *
@@ -1237,9 +1338,26 @@ get_item_tag
 	}
 } */
 
+#define INDENT(cf) \
+	if (cf->style != HTML_CLUEFLOW_STYLE_PRE) { \
+		if (! write_indent (state, cf->level)) { \
+				return FALSE; \
+		} \
+	}
+#define INDENT_T(cf) \
+	if (cf->style != HTML_CLUEFLOW_STYLE_PRE) { \
+		if (! write_indent (state, cf->level)) { \
+                        g_free (tag); \
+			return FALSE; \
+		} \
+	}
+
 static const char *
 get_tag (HTMLClueFlow *flow)
 {
+	if (!flow)
+		return NULL;
+
 	switch (flow->style) {
 	case HTML_CLUEFLOW_STYLE_LIST_ITEM:
 		return NULL;
@@ -1253,51 +1371,41 @@ get_tag (HTMLClueFlow *flow)
 	case HTML_CLUEFLOW_STYLE_ADDRESS:
 	case HTML_CLUEFLOW_STYLE_PRE:
 	default:
-		return "blockquote";
+		return "BLOCKQUOTE";
 	}
 }
 
 static gboolean
-write_indentation_tags (HTMLEngineSaveState *state,
-			guint last_value,
-			guint new_value,
-			const gchar *tag)
+write_indentation_tags (HTMLEngineSaveState *state, guint last_value, guint new_value, const gchar *tag)
 {
-	guint i, j;
+	guint i;
 
 	if (new_value == last_value)
 		return TRUE;
 
-	if (! html_engine_save_output_string (state, "\n"))
-		return FALSE;
-
 	if (new_value > last_value) {
 		for (i = last_value; i < new_value; i++) {
-			for (j = 0; j < i; j++) {
-				if (! write_indent (state))
-					return FALSE;
-			}
-
-			if (! html_engine_save_output_string (state, "<%s>\n", tag))
+			if (! write_indent (state, i + 1) || ! html_engine_save_output_string (state, "<%s>\n", tag)) {
 				return FALSE;
+			}
 		}
 	} else {
 		for (i = last_value; i > new_value; i--) {
-			if (i > 1) {
-				for (j = 0; j < i - 1; j++) {
-					if (! html_engine_save_output_string (state, "    "))
-						return FALSE;
-				}
-			}
-
-			if (! html_engine_save_output_string (state, "</%s>\n", tag))
+			if (! write_indent (state, i) || ! html_engine_save_output_string (state, "</%s>\n", tag)) {
 				return FALSE;
+			}
 		}
 	}
 
 	return TRUE;
 }
 
+inline static gint
+get_level (HTMLClueFlow *cf)
+{
+	return cf->level;
+}
+
 static gboolean
 write_pre_tags (HTMLClueFlow *self,
 		HTMLEngineSaveState *state)
@@ -1309,27 +1417,42 @@ write_pre_tags (HTMLClueFlow *self,
 	if (prev != NULL && !HTML_IS_TABLE (HTML_CLUE (self)->head)
 	    && prev->level == self->level && prev->style == self->style) {
 		if (!is_item (self) && self->style != HTML_CLUEFLOW_STYLE_PRE) {
-			return html_engine_save_output_string (state, "<br>\n");
-		} else if (self->style == HTML_CLUEFLOW_STYLE_PRE) {
-			return html_engine_save_output_string (state, "\n");
+			if (! write_indent (state, self->level))
+				return FALSE;
+			return html_engine_save_output_string (state, "<BR>\n");
 		} else
 			return TRUE;
 	}
 
-	if (prev != NULL)
-		prev_tag = get_tag (prev);
-	else
-		prev_tag = NULL;
-
+	prev_tag = get_tag (prev);
 	curr_tag = get_tag (self);
 
 	if ((prev_tag != NULL) && (curr_tag != NULL) && (strcmp (prev_tag, curr_tag) == 0)) {
-		write_indentation_tags (state, prev->level, self->level, prev_tag);
+		write_indentation_tags (state, get_level (prev), get_level (self), prev_tag);
 	} else {
-		if (prev_tag != NULL)
-			write_indentation_tags (state, prev->level, 0, prev_tag);
-		if (curr_tag != NULL)
-			write_indentation_tags (state, 0, self->level, curr_tag);
+		if (prev_tag != NULL) {
+			if (is_item (self)) {
+				write_indentation_tags (state, get_level (prev), self->level > 0 ? self->level - 1 : 0,
+							prev_tag);
+			} else {
+				write_indentation_tags (state, get_level (prev), 0, prev_tag);
+			}
+		}
+		if (curr_tag != NULL) {
+			if (prev && is_item (prev)) {
+				write_indentation_tags (state, prev->level > 0 ? prev->level - 1 : 0,
+							get_level (self), curr_tag);
+			} else {
+				write_indentation_tags (state, 0, get_level (self), curr_tag);
+			}
+		}
+		if (curr_tag == NULL && prev_tag == NULL && prev && is_item (prev) && is_item (self)
+		    && abs (prev->level - self->level) > 1) {
+			write_indentation_tags (state,
+						prev->level < self->level ? prev->level : prev->level - 1,
+						prev->level < self->level ? self->level - 1 : self->level,
+						"BLOCKQUOTE");
+		}
 	}
 
 	return TRUE;
@@ -1346,7 +1469,9 @@ write_post_tags (HTMLClueFlow *self,
 
 	tag = get_tag (self);
 	if (tag)
-		write_indentation_tags (state, self->level, 0, tag);
+		write_indentation_tags (state, get_level (self), 0, tag);
+	else if (is_item (self) && self->level > 0)
+		write_indentation_tags (state, self->level - 1, 0, "BLOCKQUOTE");
 
 	return TRUE;
 }
@@ -1369,7 +1494,8 @@ need_list_begin (HTMLObject *self)
 	return !items_are_relative (self->prev, self)
 		&& (!self->prev
 		    || (HTML_IS_CLUEFLOW (self->prev)
-			&& HTML_CLUEFLOW (self->prev)->level < HTML_CLUEFLOW (self)->level))
+			&& (HTML_CLUEFLOW (self->prev)->level < HTML_CLUEFLOW (self)->level
+			    || !is_item (HTML_CLUEFLOW (self->prev)))))
 		    && HTML_CLUEFLOW (self)->level > 0;
 }
 
@@ -1379,56 +1505,70 @@ need_list_end (HTMLObject *self)
 	return !items_are_relative (self, self->next)
 		&& (!self->next
 		    || (HTML_IS_CLUEFLOW (self->next)
-			&& HTML_CLUEFLOW (self->next)->level < HTML_CLUEFLOW (self)->level))
+			&& (HTML_CLUEFLOW (self->next)->level < HTML_CLUEFLOW (self)->level
+			    || !is_item (HTML_CLUEFLOW (self->next)))))
 		    && HTML_CLUEFLOW (self)->level > 0;
 }
 
+static gchar *
+get_list_start_tag (HTMLObject *self)
+{
+	switch (HTML_CLUEFLOW (self)->item_type) {
+	case HTML_LIST_TYPE_UNORDERED:
+	case HTML_LIST_TYPE_MENU:
+	case HTML_LIST_TYPE_DIR:
+		return g_strdup ("LI");
+	case HTML_LIST_TYPE_ORDERED_ARABIC:
+		return g_strdup_printf ("LI TYPE=1 VALUE=%d", HTML_CLUEFLOW (self)->item_number);
+	case HTML_LIST_TYPE_ORDERED_UPPER_ROMAN:
+		return g_strdup_printf ("LI TYPE=I VALUE=%d", HTML_CLUEFLOW (self)->item_number);
+	case HTML_LIST_TYPE_ORDERED_LOWER_ROMAN:
+		return g_strdup_printf ("LI TYPE=i VALUE=%d", HTML_CLUEFLOW (self)->item_number);
+	case HTML_LIST_TYPE_ORDERED_UPPER_ALPHA:
+		return g_strdup_printf ("LI TYPE=A VALUE=%d", HTML_CLUEFLOW (self)->item_number);
+	case HTML_LIST_TYPE_ORDERED_LOWER_ALPHA:
+		return g_strdup_printf ("LI TYPE=a VALUE=%d", HTML_CLUEFLOW (self)->item_number);
+	}
+
+	return NULL;
+}
+
 static gchar *
 get_start_tag (HTMLObject *self)
 {
 	switch (HTML_CLUEFLOW (self)->style) {
 	case HTML_CLUEFLOW_STYLE_H1:
-		return g_strdup ("h1");
+		return g_strdup ("H1");
 	case HTML_CLUEFLOW_STYLE_H2:
-		return g_strdup ("h2");
+		return g_strdup ("H2");
 	case HTML_CLUEFLOW_STYLE_H3:
-		return g_strdup ("h3");
+		return g_strdup ("H3");
 	case HTML_CLUEFLOW_STYLE_H4:
-		return g_strdup ("h4");
+		return g_strdup ("H4");
 	case HTML_CLUEFLOW_STYLE_H5:
-		return g_strdup ("h5");
+		return g_strdup ("H5");
 	case HTML_CLUEFLOW_STYLE_H6:
-		return g_strdup ("h6");
+		return g_strdup ("H6");
 	case HTML_CLUEFLOW_STYLE_ADDRESS:
-		return g_strdup ("address");
+		return g_strdup ("ADDRESS");
 	case HTML_CLUEFLOW_STYLE_PRE:
-		return g_strdup ("pre");
+		return g_strdup ("PRE");
 	case HTML_CLUEFLOW_STYLE_LIST_ITEM:
 		switch (HTML_CLUEFLOW (self)->item_type) {
 		case HTML_LIST_TYPE_UNORDERED:
 		case HTML_LIST_TYPE_MENU:
 		case HTML_LIST_TYPE_DIR:
-			return g_strdup (need_list_begin (self) ? "ul>\n<li" : "li");
-		case HTML_LIST_TYPE_ORDERED_ARABIC:
-			return need_list_begin (self)
-				? g_strdup ("ol type=1>\n<li")
-				: g_strdup_printf ("li type=1 value=%d", HTML_CLUEFLOW (self)->item_number);
-		case HTML_LIST_TYPE_ORDERED_UPPER_ROMAN:
-			return need_list_begin (self)
-				? g_strdup ("ol type=I>\n<li")
-				: g_strdup_printf ("li type=I value=%d", HTML_CLUEFLOW (self)->item_number);
-		case HTML_LIST_TYPE_ORDERED_LOWER_ROMAN:
-			return need_list_begin (self)
-				? g_strdup ("ol type=i>\n<li")
-				: g_strdup_printf ("li type=i value=%d", HTML_CLUEFLOW (self)->item_number);
-		case HTML_LIST_TYPE_ORDERED_UPPER_ALPHA:
-			return need_list_begin (self)
-				? g_strdup ("ol type=A>\n<li")
-				: g_strdup_printf ("li type=A value=%d", HTML_CLUEFLOW (self)->item_number);
+			return g_strdup (need_list_begin (self) ? "UL" : NULL);
 		case HTML_LIST_TYPE_ORDERED_LOWER_ALPHA:
-			return need_list_begin (self)
-				? g_strdup ("ol type=a>\n<li")
-				: g_strdup_printf ("li type=a value=%d", HTML_CLUEFLOW (self)->item_number);
+			return g_strdup (need_list_begin (self) ? "OL TYPE=a" : NULL);
+		case HTML_LIST_TYPE_ORDERED_UPPER_ALPHA:
+			return g_strdup (need_list_begin (self) ? "OL TYPE=A" : NULL);
+		case HTML_LIST_TYPE_ORDERED_LOWER_ROMAN:
+			return g_strdup (need_list_begin (self) ? "OL TYPE=i" : NULL);
+		case HTML_LIST_TYPE_ORDERED_UPPER_ROMAN:
+			return g_strdup (need_list_begin (self) ? "OL TYPE=I" : NULL);
+		case HTML_LIST_TYPE_ORDERED_ARABIC:
+			return g_strdup (need_list_begin (self) ? "OL TYPE=1" : NULL);
 		}
 	case HTML_CLUEFLOW_STYLE_NORMAL:
 	default:
@@ -1441,29 +1581,29 @@ get_end_tag (HTMLObject *self)
 {
 	switch (HTML_CLUEFLOW (self)->style) {
 	case HTML_CLUEFLOW_STYLE_H1:
-		return g_strdup ("h1");
+		return g_strdup ("H1");
 	case HTML_CLUEFLOW_STYLE_H2:
-		return g_strdup ("h2");
+		return g_strdup ("H2");
 	case HTML_CLUEFLOW_STYLE_H3:
-		return g_strdup ("h3");
+		return g_strdup ("H3");
 	case HTML_CLUEFLOW_STYLE_H4:
-		return g_strdup ("h4");
+		return g_strdup ("H4");
 	case HTML_CLUEFLOW_STYLE_H5:
-		return g_strdup ("h5");
+		return g_strdup ("H5");
 	case HTML_CLUEFLOW_STYLE_H6:
-		return g_strdup ("h6");
+		return g_strdup ("H6");
 	case HTML_CLUEFLOW_STYLE_ADDRESS:
-		return g_strdup ("address");
+		return g_strdup ("ADDRESS");
 	case HTML_CLUEFLOW_STYLE_PRE:
-		return g_strdup ("pre");
+		return g_strdup ("PRE");
 	case HTML_CLUEFLOW_STYLE_LIST_ITEM:
 		switch (HTML_CLUEFLOW (self)->item_type) {
 		case HTML_LIST_TYPE_UNORDERED:
 		case HTML_LIST_TYPE_MENU:
 		case HTML_LIST_TYPE_DIR:
-			return g_strdup (need_list_end (self) ? "li>\n</ul" : "li");
+			return need_list_end (self) ? g_strdup ("UL") : NULL;
 		default:
-			return g_strdup (need_list_end (self) ? "li>\n</ol" : "li");
+			return need_list_end (self) ? g_strdup ("OL") : NULL;
 		}
 	default:
 		return NULL;
@@ -1478,7 +1618,6 @@ save (HTMLObject *self,
 	HTMLHAlignType halign;
 	gchar *tag;
 	gboolean start = TRUE, end = TRUE;
-	gint i;
 
 	clueflow = HTML_CLUEFLOW (self);
 	halign = HTML_CLUE (self)->halign;
@@ -1491,55 +1630,76 @@ save (HTMLObject *self,
 
 	if (is_similar (self, self->next))
 		end = FALSE;
-
-	/* Indentation.  */
-	if (clueflow->style != HTML_CLUEFLOW_STYLE_PRE) {
-		for (i = 0; i < clueflow->level; i++) {
-			if (! write_indent (state))
-				return FALSE;
-		}
-	}
 	
-	/* Alignment tag.  */
-	if (halign != HTML_HALIGN_NONE && halign != HTML_HALIGN_LEFT) {
-		if (! html_engine_save_output_string
-		    (state, "<div align=%s>\n",
-		     html_engine_save_get_paragraph_align (html_alignment_to_paragraph (halign))))
-			return FALSE;
-	}
+	INDENT (clueflow);
 
 	/* Start tag.  */
 	tag = get_start_tag (self);
-	if (tag != NULL && (start || is_item (clueflow))
-	    && (! html_engine_save_output_string (state, "<%s>", tag))) {
-		g_free (tag);
-		return FALSE;
+	if (start || is_item (clueflow)) {
+		if (tag && ! html_engine_save_output_string (state, "<%s>", tag)) {
+			g_free (tag);
+			return FALSE;
+		}
+		if (is_item (clueflow)) {
+			if (tag) {
+				g_free (tag);
+				if (! html_engine_save_output_string (state, "\n")) {
+					return FALSE;
+				}
+				INDENT (clueflow);
+			}
+			tag = get_list_start_tag (self);
+			if (! html_engine_save_output_string (state, "<%s>", tag)) {
+				g_free (tag);
+				return FALSE;
+			}
+		}
 	}
 	g_free (tag);
 
+	/* Alignment tag.  */
+	if (halign != HTML_HALIGN_NONE && halign != HTML_HALIGN_LEFT) {
+		if (! html_engine_save_output_string
+		    (state, "<DIV ALIGN=%s>",
+		     html_engine_save_get_paragraph_align (html_alignment_to_paragraph (halign))))
+			return FALSE;
+	}
+
 	/* Paragraph's content.  */
 	if (! HTML_OBJECT_CLASS (&html_clue_class)->save (self, state))
 		return FALSE;
 
+	/* Close alignment tag.  */
+	if (halign != HTML_HALIGN_NONE && halign != HTML_HALIGN_LEFT) {
+		if (! html_engine_save_output_string (state, "</DIV>"))
+			return FALSE;
+	}
+
 	/* End tag.  */
 	tag = get_end_tag (self);
-	if (tag && (end || is_item (clueflow))) {
-		if (! html_engine_save_output_string (state, "</%s>", tag)) {
+	if (end || is_item (clueflow)) {
+		if (is_item (clueflow)) {
+			if (! html_engine_save_output_string (state, "</LI>")) {
+				g_free (tag);
+				return FALSE;
+			}
+			if (tag) {
+				if (! html_engine_save_output_string (state, "\n")) {
+					g_free (tag);
+					return FALSE;
+				}
+				INDENT_T (clueflow);
+			}
+		}
+		if (tag && ! html_engine_save_output_string (state, "</%s>", tag)) {
 			g_free (tag);
 			return FALSE;
 		}
 	}
 	g_free (tag);
 
-	/* Close alignment tag.  */
-	if (halign != HTML_HALIGN_NONE && halign != HTML_HALIGN_LEFT) {
-		if (! html_engine_save_output_string (state, "</div>\n"))
-			return FALSE;
-	} else if (tag != NULL
-		   && HTML_CLUEFLOW (self)->style != HTML_CLUEFLOW_STYLE_PRE) {
-		if (! html_engine_save_output_string (state, "\n"))
-			return FALSE;
-	}
+	if (! html_engine_save_output_string (state, "\n"))
+		return FALSE;
 
 	return write_post_tags (HTML_CLUEFLOW (self), state);
 }
@@ -1764,7 +1924,7 @@ get_default_font_style (const HTMLClueFlow *self)
 }
 
 static void
-search_set_info (HTMLObject *cur, HTMLSearch *info, guint pos, guint len)
+search_set_info (HTMLObject *cur, HTMLSearch *info, guchar *text, guint pos, guint len)
 {
 	guint text_len = 0;
 	guint cur_len;
@@ -1778,16 +1938,17 @@ search_set_info (HTMLObject *cur, HTMLSearch *info, guint pos, guint len)
 
 	while (cur) {
 		if (html_object_is_text (cur)) {
-			cur_len = HTML_TEXT (cur)->text_len;
+			cur_len = strlen (HTML_TEXT (cur)->text);
 			if (text_len + cur_len > pos) {
 				if (!info->found) {
-					info->start_pos = pos - text_len;
+					info->start_pos = g_utf8_pointer_to_offset (text + text_len,
+										    text + pos);
 				}
 				info->found = g_list_append (info->found, cur);
 			}
 			text_len += cur_len;
 			if (text_len >= pos+info->found_len) {
-				info->stop_pos = cur_len - (text_len - pos - info->found_len);
+				info->stop_pos = info->start_pos + info->found_len;
 				info->last     = HTML_OBJECT (cur);
 				return;
 			}
@@ -1819,7 +1980,7 @@ search_text (HTMLObject **beg, HTMLSearch *info)
 	text_len = 0;
 	while (cur) {
 		if (html_object_is_text (cur)) {
-			text_len += HTML_TEXT (cur)->text_len;
+			text_len += strlen (HTML_TEXT (cur)->text);
 			end = cur;
 		} else if (HTML_OBJECT_TYPE (cur) != HTML_TYPE_TEXTSLAVE) {
 			break;
@@ -1839,11 +2000,11 @@ search_text (HTMLObject **beg, HTMLSearch *info)
 		while (cur) {
 			if (html_object_is_text (cur)) {
 				if (!info->forward) {
-					pp -= HTML_TEXT (cur)->text_len;
+					pp -= strlen (HTML_TEXT (cur)->text);
 				}
-				strncpy (pp, HTML_TEXT (cur)->text, HTML_TEXT (cur)->text_len);
+				strncpy (pp, HTML_TEXT (cur)->text, strlen (HTML_TEXT (cur)->text));
 				if (info->forward) {
-					pp += HTML_TEXT (cur)->text_len;
+					pp += strlen (HTML_TEXT (cur)->text);
 				}
 			} else if (HTML_OBJECT_TYPE (cur) != HTML_TYPE_TEXTSLAVE) {
 				break;
@@ -1872,13 +2033,14 @@ search_text (HTMLObject **beg, HTMLSearch *info)
 				regmatch_t match;
 				guchar *p=par+pos;
 
-				/* replace &nbsp;'s with spaces */
+				/* FIXME UTF8
+				   replace &nbsp;'s with spaces
 				while (*p) {
 					if (*p == ENTITY_NBSP) {
 						*p = ' ';
 					}
 					p += (info->forward) ? 1 : -1;
-				}
+					} */
 
 				while ((info->forward && pos < text_len)
 				       || (!info->forward && pos >= 0)) {
@@ -1886,7 +2048,8 @@ search_text (HTMLObject **beg, HTMLSearch *info)
 						      par + pos,
 						      1, &match, 0);
 					if (rv == 0) {
-						search_set_info (head, info, pos + match.rm_so, match.rm_eo - match.rm_so);
+						search_set_info (head, info, par,
+								 pos + match.rm_so, match.rm_eo - match.rm_so);
 						retval = TRUE;
 						break;
 					}
@@ -1901,7 +2064,7 @@ search_text (HTMLObject **beg, HTMLSearch *info)
 					if (rv < 0) {
 						g_warning ("re_match (...) error");
 					}
-					search_set_info (head, info, found_pos, rv);
+					search_set_info (head, info, par, found_pos, rv);
 					retval = TRUE;
 				} else {
 					if (rv < -1) {
@@ -1918,8 +2081,9 @@ search_text (HTMLObject **beg, HTMLSearch *info)
 					    == info->trans [par [pos]]) {
 						eq_len++;
 						if (eq_len == info->text_len) {
-							search_set_info (head, info, pos - ((info->forward)
-											    ? eq_len-1 : 0), info->text_len);
+							search_set_info (head, info, par,
+									 pos - (info->forward ? eq_len-1 : 0),
+									 info->text_len);
 							retval=TRUE;
 							break;
 						}
@@ -2002,7 +2166,7 @@ relayout (HTMLObject *self,
 	mw = html_object_calc_min_width (self, engine->painter);
 	if (mw <= self->max_width)
 		return (*HTML_OBJECT_CLASS (parent_class)->relayout) (self, engine, child);
-	html_engine_calc_size (engine);
+	html_engine_calc_size (engine, FALSE);
 	html_engine_draw (engine, 0, 0, engine->width, engine->height);
 
 	return TRUE;
@@ -2515,25 +2679,31 @@ spell_check_word_mark (HTMLObject *obj, const gchar *text, const gchar *word, gu
 }
 
 static gchar *
-begin_of_word (gchar *text, gchar *ct)
+begin_of_word (gchar *text, gchar *ct, gboolean *cited)
 {
 	gunichar uc;
 
+	*cited = FALSE;
 	do
 		uc = g_utf8_get_char (ct);
-	while (!html_is_in_word (uc) && (ct = g_utf8_next_char (ct)) && *ct);
+	while (!html_selection_spell_word (uc, cited) && (ct = g_utf8_next_char (ct)) && *ct);
 
 	return ct;
 }
 
 static gchar *
-end_of_word (gchar *ct)
+end_of_word (gchar *ct, gboolean cited)
 {
 	gunichar uc;
 	gchar *cn;
+	gboolean cited2;
 
-	while (*ct && (cn = e_unicode_get_utf8 (ct, &uc)) && html_is_in_word (uc))
+	cited2 = FALSE;
+	while (*ct && (cn = e_unicode_get_utf8 (ct, &uc))
+	       && (html_selection_spell_word (uc, &cited2) || (!cited && cited2))) {
 		ct = cn;
+		cited2 = FALSE;
+	}
 
 	return ct;
 }
@@ -2581,8 +2751,10 @@ html_clueflow_spell_check (HTMLClueFlow *flow, HTMLEngine *e, HTMLInterval *inte
 	if (text) {
 		ct = text;
 		while (*ct) {
-			word = ct = begin_of_word (text, ct);
-			ct        =   end_of_word (ct);
+			gboolean cited;
+
+			word = ct = begin_of_word (text, ct, &cited);
+			ct        =   end_of_word (ct, cited);
 
 			/* test if we have found word */
 			if (word != ct) {
